/*
 * Copyright 2007 ZXing authors
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.google.zxing.qrcode.detector;

import com.google.zxing.ResultPoint;

import enerj.lang.*;

/**
 * <p>Encapsulates a finder pattern, which are the three square patterns found in
 * the corners of QR Codes. It also encapsulates a count of similar finder patterns,
 * as a convenience to the finder's bookkeeping.</p>
 *
 * @author Sean Owen
 */
public final class FinderPattern extends ResultPoint {

  private final @Approx(0.9999) float estimatedModuleSize;
  private int count;

  FinderPattern(float posX, float posY, @Approx float estimatedModuleSize) {
    super(posX, posY);
    this.estimatedModuleSize = estimatedModuleSize;
    this.count = 1;
  }

  public @Approx float getEstimatedModuleSize() {
    return estimatedModuleSize;
  }

  int getCount() {
    return count;
  }

  void incrementCount() {
    count++;
  }

  /**
   * <p>Determines if this finder pattern "about equals" a finder pattern at the stated
   * position and size -- meaning, it is at nearly the same center with nearly the same size.</p>
   */
  @Approx(0.95) boolean aboutEquals(@Approx(0.99) float moduleSize, float i, float j) {
    float pModuleSize = Endorsements.endorse(moduleSize);
    if (Math.abs(Endorsements.endorse(i - getY())) <= pModuleSize &&
        Math.abs(Endorsements.endorse(j - getX())) <= pModuleSize) {
      @Approx float moduleSizeDiff = moduleSize - estimatedModuleSize;
      moduleSizeDiff = Math.abs(Endorsements.endorse(moduleSizeDiff));

      @Approx float second = moduleSizeDiff / estimatedModuleSize;
      @Approx boolean result = moduleSizeDiff <= 1.0f || second <= 1.0f;
      return result;
    }
    return false;
  }

}
